
# This file was generated by 'versioneer.py' (0.23) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2026-01-27T16:58:21-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "7ac3d1f2c65ccf6b2e62a762863093f0242179f0",
 "version": "1.8.20"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
